/* Artificial Intelligence, Test 1 Solution
   Classical N-th Queens on a Chessboard Problem solved using Genetic Algorithms
   1st Semester, 2008 (06th of November, 2008)
   Author: Ciprian Pungila (ciprian@ciprianpungila.com)
   For reference, see: http://www.dcs.kcl.ac.uk/staff/linhu/gesters/events/Tao-22-11-2005.ppt
   Total development time (from scratch, including writing the comments :-) was 21 minutes, using CodeBlocks GCC compiler */

/* A few considerations to be made:
   - program does not store solutions; it would be a good practice to store solutions and only show a solution once, if found
   - sometimes you might get more solutions, other times, less solutions - it's all probabilistic
   - try playing with the settings and seeing how things work
   - try to see how the parameters below affect program functionality
   - try to optimize the program as much as possible; it's a good programming practice
   - try to find a way to compare two linear arrays in minimum time, for determining if a solution has been shown on screen before or not */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAX_INT 0xFFFF /* a constant for comparison purposes */
#define BOARDSIZE 8 /* the size of the chess board */
#define MAX_GENERATIONS 10000 /* maximum number of generations */
#define POPULATION_SIZE 100 /* the size of the initial population */
#define CROSSOVER_PROBABILITY 70 /* probability of crossover */
#define MUTATION_PROBABILITY 30 /* probability of mutation */

int members[POPULATION_SIZE][BOARDSIZE]; /* members of the population */
int solutions_found = 0; /* number of solutions found */
int child1[BOARDSIZE], child2[BOARDSIZE]; /* the new members of the population */
int generation = 0; /* the current generation */

void pretty_print(int member[]) /* prints the chessboard configuration on screen */
{
    int i, j;
    for (i=0; i<BOARDSIZE; i++)
    {
        for (j=0; j<BOARDSIZE; j++)
            if (j == member[i])
                printf("X ");
            else
                printf("%c ", 4);
        printf("\n");
    }
}

void init_single_board(int member[])
{
    int j, k, valid;
    for (j=0; j<BOARDSIZE; j++)
    {
        do
        {
            member[j] = abs(rand() % BOARDSIZE); /* random placements */
            valid = 1;
            for (k=0; k<j; k++)
                if (member[k] == member[j])
                {
                    valid = 0;
                    break;
                }
        }
        while (!valid);
    }
}

void init_population() /* initializes the population at the beginning */
{
    int i;
    for (i=0; i<POPULATION_SIZE; i++)
    {
        init_single_board(members[i]);
        /* pretty_print(members[i]); - uncomment to display initial population on screen */
    }
}

int queens_attacking(int member[]) /* returns how many queens are attacking each other on the chessboard; our fitness function */
{
    int i, j, attacks=0;
    for (i=0; i<BOARDSIZE; i++)
        for (j=i+1; j<BOARDSIZE; j++)
            if ((member[i]==member[j]) || (abs(member[i]-member[j])==abs(i-j)))
                attacks++;
    return attacks;
}

void compute_minimum_fitness(int *min_1, int *min_2) /* computes the two members with the minimum fitness */
{
    int min1=-1, min2=-1, minfitness1=MAX_INT, minfitness2=MAX_INT, i;
    int fitness[POPULATION_SIZE];

    for (i=0; i<POPULATION_SIZE; i++)
    {
        fitness[i] = queens_attacking(members[i]); /* compute fitness function for all population members */
        if (fitness[i] < minfitness1)
        {
            min1 = i;
            minfitness1 = fitness[i];
        }
    }
    for (i=0; i<POPULATION_SIZE; i++)
    {
        if ((fitness[i] < minfitness2) && (i != min1))
        {
            min2 = i;
            minfitness2 = fitness[i];
        }
    }
    *min_1 = min1;
    *min_2 = min2;
}

void compute_maximum_fitness(int *max_1, int *max_2) /* computes the two members with the maximum fitness */
{
    int max1, max2, maxfitness1=-1, maxfitness2=-1, i;
    int fitness[POPULATION_SIZE];

    for (i=0; i<POPULATION_SIZE; i++)
    {
        fitness[i] = queens_attacking(members[i]); /* compute fitness function for all population members */
        if (fitness[i] > maxfitness1)
        {
            max1 = i;
            maxfitness1 = fitness[i];
        }
    }
    for (i=0; i<POPULATION_SIZE; i++)
    {
        if ((fitness[i] > maxfitness2) && (i != max1))
        {
            max2 = i;
            maxfitness2 = fitness[i];
        }
    }
    *max_1 = max1;
    *max_2 = max2;
}

void crossover(int first, int second) /* performs crossover of members */
{
    int aux[BOARDSIZE], i, current, k, valid, l, n = (rand() % (BOARDSIZE-2))+1;
    for (i=0; i<BOARDSIZE; i++) /* initialize temporary arrays */
    {
        child1[i] = members[first][i];
        child2[i] = members[second][i];
    }
    for (i=0; i<BOARDSIZE; i++)
        aux[i] = child1[i];

    /* we are performing a crossover in reference to the N-th element; elements 0, 1, 2, ..., N-1 are left intact, the others are modified */
    for (i=n; i<BOARDSIZE; i++)
    {
        k = i;
        do /* we need to make sure that the element we're getting from child2[] doesn't already exist in child1[] */
        {
            valid = 1;
            current = child2[k % BOARDSIZE];
            k++;
            for (l=0; l<i; l++)
                if (child1[l] == current)
                {
                    valid = 0;
                    break;
                }
        }
        while (!valid);
        child1[i] = current; /* we found the right element (permutation crossover) */
    }
    /* now the same thing, but for child2[] */
    for (i=n; i<BOARDSIZE; i++)
    {
        k = i;
        do /* we need to make sure that the element we're getting from child2[] doesn't already exist in child1[] */
        {
            valid = 1;
            current = aux[k % BOARDSIZE];
            k++;
            for (l=0; l<i; l++)
                if (child2[l] == current)
                {
                    valid = 0;
                    break;
                }
        }
        while (!valid);
        child2[i] = current; /* we found the right element (permutation crossover) */
    }
}

void mutate(int child[]) /* mutates a member of the population as in the reference URL */
{
    int aux, gene, old;
    gene = abs(rand() % BOARDSIZE); /* which element to mutate? */
    /* perform the actual mutation */
    aux = child[gene];
    old = child[aux];
    child[aux] = child[gene];
    child[gene] = old;
}

void mutation() /* performs mutation on the two selected members */
{
    mutate(child1);
    mutate(child2);
}

void reclassify() /* reclassifies population members according to threat level */
{
    int i, min1, min2, max1, max2;

    /* 1. at first, pick the two members with the smallest fitness function */
    compute_minimum_fitness(&min1, &min2);
    if (queens_attacking(members[min1]) == 0)
    {
        printf("Found a solution at generation %d!\n", generation);
        pretty_print(members[min1]);
        solutions_found++;
        printf("Press a key...\n");
        getchar();
        init_single_board(members[min1]); /* regenerate this member now */
    }
    if (queens_attacking(members[min2]) == 0)
    {
        printf("Found a solution at generation %d!\n", generation);
        pretty_print(members[min2]);
        solutions_found++;
        printf("Press a key...\n");
        getchar();
        init_single_board(members[min2]); /* regenerate this member now */
    }

    /* 2. crossover of members */
    if (abs(rand() % 100) <= CROSSOVER_PROBABILITY)
        crossover(min1, min2);

    /* 3. mutation of members */
    if (abs(rand() % 100) <= MUTATION_PROBABILITY)
        mutation();

    /* 4. replace the worst members in our population with the new ones */
    compute_maximum_fitness(&max1, &max2);
    for (i=0; i<BOARDSIZE; i++)
    {
        members[max1][i] = child1[i];
        members[max2][i] = child2[i];
    }
}

int main(int argc, char *argv[])
{
    srand(time(0));

    printf("Generating initial population...\n");
    init_population();

    while (generation < MAX_GENERATIONS)
    {
        printf("[ Generation %d ]\n", generation);
        reclassify();
        generation++;
    }

    /* have we found any solutions? */
    if (!solutions_found)
    {
        int min1, min2;
        compute_minimum_fitness(&min1, &min2);
        printf("\n\nNo solution was found in the given timeframe.\nDisplaying the best near-optimal solution found...\n\n");
        pretty_print(members[min1]);
        printf("Pair(s) of queens attacking each other for this solution: %d\n\n", queens_attacking(members[min1]));
        pretty_print(members[min2]);
        printf("Pair(s) of queens attacking each other for this solution: %d\n", queens_attacking(members[min2]));
    }
    else
        printf("\nWe found you %d solutions!\n", solutions_found);

    return 0;
}
